@@ -5,26 +5,26 @@ package IkiWiki;
 use warnings;
 use strict;
 use Encode;
-use HTML::Entities;
 use URI::Escape q{uri_escape_utf8};
-use POSIX;
+use POSIX ();
 use Storable;
 use open qw{:utf8 :std};
 
 use vars qw{%config %links %oldlinks %pagemtime %pagectime %pagecase
-	    %pagestate %wikistate %renderedfiles %oldrenderedfiles
-	    %pagesources %destsources %depends %depends_simple %hooks
-	    %forcerebuild %loaded_plugins};
+	%pagestate %wikistate %renderedfiles %oldrenderedfiles
+	%pagesources %delpagesources %destsources %depends %depends_simple
+	@mass_depends %hooks %forcerebuild %loaded_plugins %typedlinks
+	%oldtypedlinks %autofiles};
 
 use Exporter q{import};
-our @EXPORT = qw(hook debug error template htmlpage deptype
-                 add_depends pagespec_match pagespec_match_list bestlink
-		 htmllink readfile writefile pagetype srcfile pagename
-		 displaytime will_render gettext urlto targetpage
-		 add_underlay pagetitle titlepage linkpage newpagefile
-		 inject add_link
-                 %config %links %pagestate %wikistate %renderedfiles
-                 %pagesources %destsources);
+our @EXPORT = qw(hook debug error htmlpage template template_depends
+	deptype add_depends pagespec_match pagespec_match_list bestlink
+	htmllink readfile writefile pagetype srcfile pagename
+	displaytime will_render gettext ngettext urlto targetpage
+	add_underlay pagetitle titlepage linkpage newpagefile
+	inject add_link add_autofile
+	%config %links %pagestate %wikistate %renderedfiles
+	%pagesources %destsources %typedlinks);
 our $VERSION = 3.00; # plugin interface version, next is ikiwiki version
 our $version='unknown'; # VERSION_AUTOREPLACE done by Makefile, DNE
 our $installdir='/usr'; # INSTALLDIR_AUTOREPLACE done by Makefile, DNE
@@ -37,6 +37,7 @@ our $DEPEND_LINKS=4;
 # Optimisation.
 use Memoize;
 memoize("abs2rel");
+memoize("sortspec_translate");
 memoize("pagespec_translate");
 memoize("template_file");
 
@@ -150,18 +151,11 @@ sub getsetup () {
 	templatedir => {
 		type => "string",
 		default => "$installdir/share/ikiwiki/templates",
-		description => "location of template files",
+		description => "additional directory to search for template files",
 		advanced => 1,
 		safe => 0, # path
 		rebuild => 1,
 	},
-	templatedirs => {
-		type => "internal",
-		default => [],
-		description => "additional directories containing template files",
-		safe => 0,
-		rebuild => 0,
-	},
 	underlaydir => {
 		type => "string",
 		default => "$installdir/share/ikiwiki/basewiki",
@@ -240,6 +234,14 @@ sub getsetup () {
 		safe => 1,
 		rebuild => 1,
 	},
+	html5 => {
+		type => "boolean",
+		default => 0,
+		description => "generate HTML5? (experimental)",
+		advanced => 1,
+		safe => 1,
+		rebuild => 1,
+	},
 	sslcookie => {
 		type => "boolean",
 		default => 0,
@@ -334,18 +336,27 @@ sub getsetup () {
 		safe => 0, # paranoia
 		rebuild => 0,
 	},
+	include => {
+		type => "string",
+		default => undef,
+		example => '^\.htaccess$',
+		description => "regexp of normally excluded files to include",
+		advanced => 1,
+		safe => 0, # regexp
+		rebuild => 1,
+	},
 	exclude => {
 		type => "string",
 		default => undef,
-		example => '\.wav$',
-		description => "regexp of source files to ignore",
+		example => '^(*\.private|Makefile)$',
+		description => "regexp of files that should be skipped",
 		advanced => 1,
 		safe => 0, # regexp
 		rebuild => 1,
 	},
 	wiki_file_prune_regexps => {
 		type => "internal",
-		default => [qr/(^|\/)\.\.(\/|$)/, qr/^\./, qr/\/\./,
+		default => [qr/(^|\/)\.\.(\/|$)/, qr/^\//, qr/^\./, qr/\/\./,
 			qr/\.x?html?$/, qr/\.ikiwiki-new$/,
 			qr/(^|\/).svn\//, qr/.arch-ids\//, qr/{arch}\//,
 			qr/(^|\/)_MTN\//, qr/(^|\/)_darcs\//,
@@ -409,6 +420,13 @@ sub getsetup () {
 		safe => 0,
 		rebuild => 0,
 	},
+	clean => {
+		type => "internal",
+		default => 0,
+		description => "running in clean mode",
+		safe => 0,
+		rebuild => 0,
+	},
 	refresh => {
 		type => "internal",
 		default => 0,
@@ -423,10 +441,16 @@ sub getsetup () {
 		safe => 0,
 		rebuild => 0,
 	},
-	getctime => {
+	wrapper_background_command => {
 		type => "internal",
-		default => 0,
-		description => "running in getctime mode",
+		default => '',
+		description => "background shell command to run",
+		safe => 0,
+		rebuild => 0,
+	},
+	gettime => {
+		type => "internal",
+		description => "running in gettime mode",
 		safe => 0,
 		rebuild => 0,
 	},
@@ -451,6 +475,13 @@ sub getsetup () {
 		safe => 0,
 		rebuild => 0,
 	},
+	setuptype => {
+		type => "internal",
+		default => "Standard",
+		description => "perl class to use to dump setup file",
+		safe => 0,
+		rebuild => 0,
+	},
 	allow_symlinks_before_srcdir => {
 		type => "boolean",
 		default => 0,
@@ -568,10 +599,11 @@ sub loadplugins () {
 	return 1;
 }
 
-sub loadplugin ($) {
+sub loadplugin ($;$) {
 	my $plugin=shift;
+	my $force=shift;
 
-	return if grep { $_ eq $plugin} @{$config{disable_plugins}};
+	return if ! $force && grep { $_ eq $plugin} @{$config{disable_plugins}};
 
 	foreach my $dir (defined $config{libdir} ? possibly_foolish_untaint($config{libdir}) : undef,
 	                 "$installdir/lib/ikiwiki") {
@@ -791,6 +823,17 @@ sub prep_writefile ($$) {
 		if (-l "$destdir/$test") {
 			error("cannot write to a symlink ($test)");
 		}
+		if (-f _ && $test ne $file) {
+			# Remove conflicting file.
+			foreach my $p (keys %renderedfiles, keys %oldrenderedfiles) {
+				foreach my $f (@{$renderedfiles{$p}}, @{$oldrenderedfiles{$p}}) {
+					if ($f eq $test) {
+						unlink("$destdir/$test");
+						last;
+					}
+				}
+			}
+		}
 		$test=dirname($test);
 	}
 
@@ -844,10 +887,36 @@ sub will_render ($$;$) {
 	my $dest=shift;
 	my $clear=shift;
 
-	# Important security check.
+	# Important security check for independently created files.
 	if (-e "$config{destdir}/$dest" && ! $config{rebuild} &&
 	    ! grep { $_ eq $dest } (@{$renderedfiles{$page}}, @{$oldrenderedfiles{$page}}, @{$wikistate{editpage}{previews}})) {
-		error("$config{destdir}/$dest independently created, not overwriting with version from $page");
+		my $from_other_page=0;
+	    	# Expensive, but rarely runs.
+		foreach my $p (keys %renderedfiles, keys %oldrenderedfiles) {
+			if (grep {
+				$_ eq $dest ||
+				dirname($_) eq $dest
+			    } @{$renderedfiles{$p}}, @{$oldrenderedfiles{$p}}) {
+				$from_other_page=1;
+				last;
+			}
+		}
+
+		error("$config{destdir}/$dest independently created, not overwriting with version from $page")
+			unless $from_other_page;
+	}
+
+	# If $dest exists as a directory, remove conflicting files in it
+	# rendered from other pages.
+	if (-d _) {
+		foreach my $p (keys %renderedfiles, keys %oldrenderedfiles) {
+			foreach my $f (@{$renderedfiles{$p}}, @{$oldrenderedfiles{$p}}) {
+				if (dirname($f) eq $dest) {
+					unlink("$config{destdir}/$f");
+					rmdir(dirname("$config{destdir}/$f"));
+				}
+			}
+		}
 	}
 
 	if (! $clear || $cleared{$page}) {
@@ -941,7 +1010,12 @@ sub linkpage ($) {
 sub cgiurl (@) {
 	my %params=@_;
 
-	return $config{cgiurl}."?".
+	my $cgiurl=$config{cgiurl};
+	if (exists $params{cgiurl}) {
+		$cgiurl=$params{cgiurl};
+		delete $params{cgiurl};
+	}
+	return $cgiurl."?".
 		join("&amp;", map $_."=".uri_escape_utf8($params{$_}), keys %params);
 }
 
@@ -969,10 +1043,18 @@ sub abs2rel ($$) {
 	return $ret;
 }
 
-sub displaytime ($;$) {
+sub displaytime ($;$$) {
 	# Plugins can override this function to mark up the time to
 	# display.
-	return '<span class="date">'.formattime(@_).'</span>';
+	my $time=formattime($_[0], $_[1]);
+	if ($config{html5}) {
+		return '<time datetime="'.date_3339($_[0]).'"'.
+			($_[2] ? ' pubdate="pubdate"' : '').
+			'>'.$time.'</time>';
+	}
+	else {
+		return '<span class="date">'.$time.'</span>';
+	}
 }
 
 sub formattime ($;$) {
@@ -988,6 +1070,16 @@ sub formattime ($;$) {
 	return decode_utf8(POSIX::strftime($format, localtime($time)));
 }
 
+sub date_3339 ($) {
+	my $time=shift;
+
+	my $lc_time=POSIX::setlocale(&POSIX::LC_TIME);
+	POSIX::setlocale(&POSIX::LC_TIME, "C");
+	my $ret=POSIX::strftime("%Y-%m-%dT%H:%M:%SZ", gmtime($time));
+	POSIX::setlocale(&POSIX::LC_TIME, $lc_time);
+	return $ret;
+}
+
 sub beautify_urlpath ($) {
 	my $url=shift;
 
@@ -1026,6 +1118,15 @@ sub urlto ($$;$) {
 	return beautify_urlpath($link);
 }
 
+sub isselflink ($$) {
+	# Plugins can override this function to support special types
+	# of selflinks.
+	my $page=shift;
+	my $link=shift;
+
+        return $page eq $link;
+}
+
 sub htmllink ($$$;@) {
 	my $lpage=shift; # the page doing the linking
 	my $page=shift; # the page that will contain the link (different for inline)
@@ -1051,21 +1152,23 @@ sub htmllink ($$$;@) {
 	}
 	
 	return "<span class=\"selflink\">$linktext</span>"
-		if length $bestlink && $page eq $bestlink &&
+		if length $bestlink && isselflink($page, $bestlink) &&
 		   ! defined $opts{anchor};
 	
 	if (! $destsources{$bestlink}) {
 		$bestlink=htmlpage($bestlink);
 
 		if (! $destsources{$bestlink}) {
-			return $linktext unless length $config{cgiurl};
-			return "<span class=\"createlink\"><a href=\"".
-				cgiurl(
-					do => "create",
-					page => lc($link),
-					from => $lpage
-				).
-				"\" rel=\"nofollow\">?</a>$linktext</span>"
+			my $cgilink = "";
+			if (length $config{cgiurl}) {
+				$cgilink = "<a href=\"".
+					cgiurl(
+						do => "create",
+						page => lc($link),
+						from => $lpage
+					)."\" rel=\"nofollow\">?</a>";
+			}
+			return "<span class=\"createlink\">$cgilink$linktext</span>"
 		}
 	}
 	
@@ -1090,6 +1193,11 @@ sub htmllink ($$$;@) {
 	return "<a href=\"$bestlink\"@attrs>$linktext</a>";
 }
 
+sub userpage ($) {
+	my $user=shift;
+	return length $config{userdir} ? "$config{userdir}/$user" : $user;
+}
+
 sub openiduser ($) {
 	my $user=shift;
 
@@ -1098,11 +1206,10 @@ sub openiduser ($) {
 		my $display;
 
 		if (Net::OpenID::VerifiedIdentity->can("DisplayOfURL")) {
-			# this works in at least 2.x
 			$display = Net::OpenID::VerifiedIdentity::DisplayOfURL($user);
 		}
 		else {
-			# this only works in 1.x
+			# backcompat with old version
 			my $oid=Net::OpenID::VerifiedIdentity->new(identity => $user);
 			$display=$oid->display;
 		}
@@ -1115,7 +1222,7 @@ sub openiduser ($) {
 		# Convert "http://somehost.com/user" to "user [somehost.com]".
 		# (also "https://somehost.com/user/")
 		if ($display !~ /\[/) {
-			$display=~s/^https?:\/\/(.+)\/([^\/]+)\/?$/$2 [$1]/;
+			$display=~s/^https?:\/\/(.+)\/([^\/#?]+)\/?(?:[#?].*)?$/$2 [$1]/;
 		}
 		$display=~s!^https?://!!; # make sure this is removed
 		eval q{use CGI 'escapeHTML'};
@@ -1125,23 +1232,6 @@ sub openiduser ($) {
 	return;
 }
 
-sub userlink ($) {
-	my $user=shift;
-
-	my $oiduser=eval { openiduser($user) };
-	if (defined $oiduser) {
-		return "<a href=\"$user\">$oiduser</a>";
-	}
-	else {
-		eval q{use CGI 'escapeHTML'};
-		error($@) if $@;
-
-		return htmllink("", "", escapeHTML(
-			length $config{userdir} ? $config{userdir}."/".$user : $user
-		), noimageinline => 1);
-	}
-}
-
 sub htmlize ($$$$) {
 	my $page=shift;
 	my $destpage=shift;
@@ -1149,7 +1239,7 @@ sub htmlize ($$$$) {
 	my $content=shift;
 	
 	my $oneline = $content !~ /\n/;
-
+	
 	if (exists $hooks{htmlize}{$type}) {
 		$content=$hooks{htmlize}{$type}{call}->(
 			page => $page,
@@ -1170,10 +1260,9 @@ sub htmlize ($$$$) {
 	
 	if ($oneline) {
 		# hack to get rid of enclosing junk added by markdown
-		# and other htmlizers
+		# and other htmlizers/sanitizers
 		$content=~s/^<p>//i;
-		$content=~s/<\/p>$//i;
-		chomp $content;
+		$content=~s/<\/p>\n*$//i;
 	}
 
 	return $content;
@@ -1228,7 +1317,7 @@ sub preprocess ($$$;$$) {
 				(?:
 					"""(.*?)"""	# 2: triple-quoted value
 				|
-					"([^"]+)"	# 3: single-quoted value
+					"([^"]*?)"	# 3: single-quoted value
 				|
 					(\S+)		# 4: unquoted value
 				)
@@ -1314,7 +1403,7 @@ sub preprocess ($$$;$$) {
 					(?:
 						""".*?"""	# triple-quoted value
 						|
-						"[^"]+"		# single-quoted value
+						"[^"]*?"	# single-quoted value
 						|
 						[^"\s\]]+	# unquoted value
 					)
@@ -1337,7 +1426,7 @@ sub preprocess ($$$;$$) {
 					(?:
 						""".*?"""	# triple-quoted value
 						|
-						"[^"]+"		# single-quoted value
+						"[^"]*?"	# single-quoted value
 						|
 						[^"\s\]]+	# unquoted value
 					)
@@ -1366,10 +1455,6 @@ sub filter ($$$) {
 	return $content;
 }
 
-sub indexlink () {
-	return "<a href=\"$config{url}\">$config{wikiname}</a>";
-}
-
 sub check_canedit ($$$;$) {
 	my $page=shift;
 	my $q=shift;
@@ -1488,7 +1573,7 @@ sub loadindex () {
 	if (! $config{rebuild}) {
 		%pagesources=%pagemtime=%oldlinks=%links=%depends=
 		%destsources=%renderedfiles=%pagecase=%pagestate=
-		%depends_simple=();
+		%depends_simple=%typedlinks=%oldtypedlinks=();
 	}
 	my $in;
 	if (! open ($in, "<", "$config{wikistatedir}/indexdb")) {
@@ -1497,6 +1582,7 @@ sub loadindex () {
 			open ($in, "<", "$config{wikistatedir}/indexdb") || return;
 		}
 		else {
+			$config{gettime}=1; # first build
 			return;
 		}
 	}
@@ -1520,8 +1606,8 @@ sub loadindex () {
 		my $d=$pages->{$src};
 		my $page=pagename($src);
 		$pagectime{$page}=$d->{ctime};
+		$pagesources{$page}=$src;
 		if (! $config{rebuild}) {
-			$pagesources{$page}=$src;
 			$pagemtime{$page}=$d->{mtime};
 			$renderedfiles{$page}=$d->{dest};
 			if (exists $d->{links} && ref $d->{links}) {
@@ -1554,6 +1640,14 @@ sub loadindex () {
 			if (exists $d->{state}) {
 				$pagestate{$page}=$d->{state};
 			}
+			if (exists $d->{typedlinks}) {
+				$typedlinks{$page}=$d->{typedlinks};
+
+				while (my ($type, $links) = each %{$typedlinks{$page}}) {
+					next unless %$links;
+					$oldtypedlinks{$page}{$type} = {%$links};
+				}
+			}
 		}
 		$oldrenderedfiles{$page}=[@{$d->{dest}}];
 	}
@@ -1602,6 +1696,10 @@ sub saveindex () {
 			$index{page}{$src}{depends_simple} = $depends_simple{$page};
 		}
 
+		if (exists $typedlinks{$page} && %{$typedlinks{$page}}) {
+			$index{page}{$src}{typedlinks} = $typedlinks{$page};
+		}
+
 		if (exists $pagestate{$page}) {
 			foreach my $id (@hookids) {
 				foreach my $key (keys %{$pagestate{$page}{$id}}) {
@@ -1629,60 +1727,124 @@ sub saveindex () {
 }
 
 sub template_file ($) {
-	my $template=shift;
-
-	foreach my $dir ($config{templatedir}, @{$config{templatedirs}},
-	                 "$installdir/share/ikiwiki/templates") {
-		return "$dir/$template" if -e "$dir/$template";
+	my $name=shift;
+	
+	my $tpage=($name =~ s/^\///) ? $name : "templates/$name";
+	if ($name !~ /\.tmpl$/ && exists $pagesources{$tpage}) {
+		$tpage=$pagesources{$tpage};
+		$name.=".tmpl";
 	}
+
+	my $template=srcfile($tpage, 1);
+	if (defined $template) {
+		return $template, $tpage, 1 if wantarray;
+		return $template;
+	}
+	else {
+		$name=~s:/::; # avoid path traversal
+		foreach my $dir ($config{templatedir},
+		                 "$installdir/share/ikiwiki/templates") {
+			if (-e "$dir/$name") {
+				$template="$dir/$name";
+				last;
+			}
+		}
+		if (defined $template) {	
+			return $template, $tpage if wantarray;
+			return $template;
+		}
+	}
+
 	return;
 }
 
-sub template_params (@) {
-	my $filename=template_file(shift);
-
-	if (! defined $filename) {
-		return if wantarray;
-		return "";
+sub template_depends ($$;@) {
+	my $name=shift;
+	my $page=shift;
+	
+	my ($filename, $tpage, $untrusted)=template_file($name);
+	if (defined $page && defined $tpage) {
+		add_depends($page, $tpage);
 	}
 
-	my @ret=(
+	return unless defined $filename;
+
+	my @opts=(
 		filter => sub {
 			my $text_ref = shift;
 			${$text_ref} = decode_utf8(${$text_ref});
 		},
-		filename => $filename,
 		loop_context_vars => 1,
 		die_on_bad_params => 0,
-		@_
+		filename => $filename,
+		@_,
+		($untrusted ? (no_includes => 1) : ()),
 	);
-	return wantarray ? @ret : {@ret};
+	return @opts if wantarray;
+
+	require HTML::Template;
+	return HTML::Template->new(@opts);
 }
 
 sub template ($;@) {
-	require HTML::Template;
-	return HTML::Template->new(template_params(@_));
+	template_depends(shift, undef, @_);
 }
 
 sub misctemplate ($$;@) {
 	my $title=shift;
-	my $pagebody=shift;
+	my $content=shift;
+	my %params=@_;
 	
-	my $template=template("misc.tmpl");
-	$template->param(
-		title => $title,
-		indexlink => indexlink(),
-		wikiname => $config{wikiname},
-		pagebody => $pagebody,
-		baseurl => baseurl(),
-		@_,
-	);
+	my $template=template("page.tmpl");
+
+	my $page="";
+	if (exists $params{page}) {
+		$page=delete $params{page};
+	}
 	run_hooks(pagetemplate => sub {
-		shift->(page => "", destpage => "", template => $template);
+		shift->(
+			page => $page,
+			destpage => $page,
+			template => $template,
+		);
 	});
+	templateactions($template, "");
+
+	$template->param(
+		dynamic => 1,
+		title => $title,
+		wikiname => $config{wikiname},
+		content => $content,
+		baseurl => baseurl(),
+		html5 => $config{html5},
+		%params,
+	);
+	
 	return $template->output;
 }
 
+sub templateactions ($$) {
+	my $template=shift;
+	my $page=shift;
+
+	my $have_actions=0;
+	my @actions;
+	run_hooks(pageactions => sub {
+		push @actions, map { { action => $_ } } 
+			grep { defined } shift->(page => $page);
+	});
+	$template->param(actions => \@actions);
+
+	if ($config{cgiurl} && exists $hooks{auth}) {
+		$template->param(prefsurl => cgiurl(do => "prefs"));
+		$have_actions=1;
+	}
+
+	if ($have_actions || @actions) {
+		$template->param(have_actions => 1);
+	}
+}
+
 sub hook (@) {
 	my %param=@_;
 	
@@ -1731,11 +1893,11 @@ sub rcs_prepedit ($) {
 	$hooks{rcs}{rcs_prepedit}{call}->(@_);
 }
 
-sub rcs_commit ($$$;$$) {
+sub rcs_commit (@) {
 	$hooks{rcs}{rcs_commit}{call}->(@_);
 }
 
-sub rcs_commit_staged ($$$) {
+sub rcs_commit_staged (@) {
 	$hooks{rcs}{rcs_commit_staged}{call}->(@_);
 }
 
@@ -1763,6 +1925,10 @@ sub rcs_getctime ($) {
 	$hooks{rcs}{rcs_getctime}{call}->(@_);
 }
 
+sub rcs_getmtime ($) {
+	$hooks{rcs}{rcs_getmtime}{call}->(@_);
+}
+
 sub rcs_receive () {
 	$hooks{rcs}{rcs_receive}{call}->();
 }
@@ -1781,14 +1947,16 @@ sub add_depends ($$;$) {
 
 	# Add explicit dependencies for influences.
 	my $sub=pagespec_translate($pagespec);
-	return if $@;
+	return unless defined $sub;
 	foreach my $p (keys %pagesources) {
 		my $r=$sub->($p, location => $page);
 		my $i=$r->influences;
+		my $static=$r->influences_static;
 		foreach my $k (keys %$i) {
+			next unless $r || $static || $k eq $page;
 			$depends_simple{$page}{lc $k} |= $i->{$k};
 		}
-		last if $r->influences_static;
+		last if $static;
 	}
 
 	$depends{$page}{$pagespec} |= $deptype;
@@ -1811,44 +1979,56 @@ sub deptype (@) {
 	return $deptype;
 }
 
-sub file_pruned ($;$) {
+my $file_prune_regexp;
+sub file_pruned ($) {
 	my $file=shift;
-	if (@_) {
-		require File::Spec;
-		$file=File::Spec->canonpath($file);
-		my $base=File::Spec->canonpath(shift);
-		return if $file eq $base;
-		$file =~ s#^\Q$base\E/+##;
+
+	if (defined $config{include} && length $config{include}) {
+		return 0 if $file =~ m/$config{include}/;
 	}
 
-	my $regexp='('.join('|', @{$config{wiki_file_prune_regexps}}).')';
-	return $file =~ m/$regexp/;
+	if (! defined $file_prune_regexp) {
+		$file_prune_regexp='('.join('|', @{$config{wiki_file_prune_regexps}}).')';
+		$file_prune_regexp=qr/$file_prune_regexp/;
+	}
+	return $file =~ m/$file_prune_regexp/;
 }
 
 sub define_gettext () {
 	# If translation is needed, redefine the gettext function to do it.
 	# Otherwise, it becomes a quick no-op.
-	no warnings 'redefine';
+	my $gettext_obj;
+	my $getobj;
 	if ((exists $ENV{LANG} && length $ENV{LANG}) ||
 	    (exists $ENV{LC_ALL} && length $ENV{LC_ALL}) ||
 	    (exists $ENV{LC_MESSAGES} && length $ENV{LC_MESSAGES})) {
-	    	*gettext=sub {
-			my $gettext_obj=eval q{
+	    	$getobj=sub {
+			$gettext_obj=eval q{
 				use Locale::gettext q{textdomain};
 				Locale::gettext->domain('ikiwiki')
 			};
-
-			if ($gettext_obj) {
-				$gettext_obj->get(shift);
-			}
-			else {
-				return shift;
-			}
 		};
 	}
-	else {
-		*gettext=sub { return shift };
-	}
+
+	no warnings 'redefine';
+	*gettext=sub {
+		$getobj->() if $getobj;
+		if ($gettext_obj) {
+			$gettext_obj->get(shift);
+		}
+		else {
+			return shift;
+		}
+	};
+	*ngettext=sub {
+		$getobj->() if $getobj;
+		if ($gettext_obj) {
+			$gettext_obj->nget(@_);
+		}
+		else {
+			return ($_[2] == 1 ? $_[0] : $_[1])
+		}
+	};
 }
 
 sub gettext {
@@ -1856,6 +2036,11 @@ sub gettext {
 	gettext(@_);
 }
 
+sub ngettext {
+	define_gettext();
+	ngettext(@_);
+}
+
 sub yesno ($) {
 	my $val=shift;
 
@@ -1887,12 +2072,91 @@ sub inject {
 	use warnings;
 }
 
-sub add_link ($$) {
+sub add_link ($$;$) {
 	my $page=shift;
 	my $link=shift;
+	my $type=shift;
 
 	push @{$links{$page}}, $link
 		unless grep { $_ eq $link } @{$links{$page}};
+
+	if (defined $type) {
+		$typedlinks{$page}{$type}{$link} = 1;
+	}
+}
+
+sub add_autofile ($$$) {
+	my $file=shift;
+	my $plugin=shift;
+	my $generator=shift;
+	
+	$autofiles{$file}{plugin}=$plugin;
+	$autofiles{$file}{generator}=$generator;
+}
+
+sub sortspec_translate ($$) {
+	my $spec = shift;
+	my $reverse = shift;
+
+	my $code = "";
+	my @data;
+	while ($spec =~ m{
+		\s*
+		(-?)		# group 1: perhaps negated
+		\s*
+		(		# group 2: a word
+			\w+\([^\)]*\)	# command(params)
+			|
+			[^\s]+		# or anything else
+		)
+		\s*
+	}gx) {
+		my $negated = $1;
+		my $word = $2;
+		my $params = undef;
+
+		if ($word =~ m/^(\w+)\((.*)\)$/) {
+			# command with parameters
+			$params = $2;
+			$word = $1;
+		}
+		elsif ($word !~ m/^\w+$/) {
+			error(sprintf(gettext("invalid sort type %s"), $word));
+		}
+
+		if (length $code) {
+			$code .= " || ";
+		}
+
+		if ($negated) {
+			$code .= "-";
+		}
+
+		if (exists $IkiWiki::SortSpec::{"cmp_$word"}) {
+			if (defined $params) {
+				push @data, $params;
+				$code .= "IkiWiki::SortSpec::cmp_$word(\$data[$#data])";
+			}
+			else {
+				$code .= "IkiWiki::SortSpec::cmp_$word(undef)";
+			}
+		}
+		else {
+			error(sprintf(gettext("unknown sort type %s"), $word));
+		}
+	}
+
+	if (! length $code) {
+		# undefined sorting method... sort arbitrarily
+		return sub { 0 };
+	}
+
+	if ($reverse) {
+		$code="-($code)";
+	}
+
+	no warnings;
+	return eval 'sub { '.$code.' }';
 }
 
 sub pagespec_translate ($) {
@@ -1962,7 +2226,7 @@ sub pagespec_match ($$;@) {
 
 	my $sub=pagespec_translate($spec);
 	return IkiWiki::ErrorReason->new("syntax error in pagespec \"$spec\"")
-		if $@ || ! defined $sub;
+		if ! defined $sub;
 	return $sub->($page, @params);
 }
 
@@ -1980,7 +2244,9 @@ sub pagespec_match_list ($$;@) {
 
 	my $sub=pagespec_translate($pagespec);
 	error "syntax error in pagespec \"$pagespec\""
-		if $@ || ! defined $sub;
+		if ! defined $sub;
+	my $sort=sortspec_translate($params{sort}, $params{reverse})
+		if defined $params{sort};
 
 	my @candidates;
 	if (exists $params{list}) {
@@ -1993,39 +2259,19 @@ sub pagespec_match_list ($$;@) {
 			? grep { ! $params{filter}->($_) } keys %pagesources
 			: keys %pagesources;
 	}
-
-	if (defined $params{sort}) {
-		my $f;
-		if ($params{sort} eq 'title') {
-			$f=sub { pagetitle(basename($a)) cmp pagetitle(basename($b)) };
-		}
-		elsif ($params{sort} eq 'title_natural') {
-			eval q{use Sort::Naturally};
-			if ($@) {
-				error(gettext("Sort::Naturally needed for title_natural sort"));
-			}
-			$f=sub { Sort::Naturally::ncmp(pagetitle(basename($a)), pagetitle(basename($b))) };
-                }
-		elsif ($params{sort} eq 'mtime') {
-			$f=sub { $pagemtime{$b} <=> $pagemtime{$a} };
-		}
-		elsif ($params{sort} eq 'age') {
-			$f=sub { $pagectime{$b} <=> $pagectime{$a} };
-		}
-		else {
-			error sprintf(gettext("unknown sort type %s"), $params{sort});
-		}
-		@candidates = sort { &$f } @candidates;
-	}
-
-	@candidates=reverse(@candidates) if $params{reverse};
-	
-	$depends{$page}{$pagespec} |= ($params{deptype} || $DEPEND_CONTENT);
 	
 	# clear params, remainder is passed to pagespec
+	$depends{$page}{$pagespec} |= ($params{deptype} || $DEPEND_CONTENT);
 	my $num=$params{num};
 	delete @params{qw{num deptype reverse sort filter list}};
 	
+	# when only the top matches will be returned, it's efficient to
+	# sort before matching to pagespec,
+	if (defined $num && defined $sort) {
+		@candidates=IkiWiki::SortSpec::sort_pages(
+			$sort, @candidates);
+	}
+	
 	my @matches;
 	my $firstfail;
 	my $count=0;
@@ -2034,6 +2280,9 @@ sub pagespec_match_list ($$;@) {
 		my $r=$sub->($p, %params, location => $page);
 		error(sprintf(gettext("cannot match pages: %s"), $r))
 			if $r->isa("IkiWiki::ErrorReason");
+		unless ($r || $r->influences_static) {
+			$r->remove_influence($p);
+		}
 		$accum |= $r;
 		if ($r) {
 			push @matches, $p;
@@ -2047,14 +2296,21 @@ sub pagespec_match_list ($$;@) {
 		$depends_simple{$page}{lc $k} |= $i->{$k};
 	}
 
-	return @matches;
+	# when all matches will be returned, it's efficient to
+	# sort after matching
+	if (! defined $num && defined $sort) {
+		return IkiWiki::SortSpec::sort_pages(
+			$sort, @matches);
+	}
+	else {
+		return @matches;
+	}
 }
 
 sub pagespec_valid ($) {
 	my $spec=shift;
 
-	my $sub=pagespec_translate($spec);
-	return ! $@;
+	return defined pagespec_translate($spec);
 }
 
 sub glob2re ($) {
@@ -2112,7 +2368,7 @@ sub merge_influences {
 	my $anded=shift;
 
 	if (! $anded || (($this || %{$this->[1]}) &&
-	                ($other || %{$other->[1]}))) {
+	                 ($other || %{$other->[1]}))) {
 		foreach my $influence (keys %{$other->[1]}) {
 			$this->[1]{$influence} |= $other->[1]{$influence};
 		}
@@ -2123,6 +2379,13 @@ sub merge_influences {
 	}
 }
 
+sub remove_influence {
+	my $this=shift;
+	my $torm=shift;
+
+	delete $this->[1]{$torm};
+}
+
 package IkiWiki::ErrorReason;
 
 our @ISA = 'IkiWiki::FailReason';
@@ -2136,7 +2399,7 @@ sub derel ($$) {
 	if ($path =~ m!^\./!) {
 		$from=~s#/?[^/]+$## if defined $from;
 		$path=~s#^\./##;
-		$path="$from/$path" if length $from;
+		$path="$from/$path" if defined $from && length $from;
 	}
 
 	return $path;
@@ -2164,7 +2427,19 @@ sub match_glob ($$;@) {
 }
 
 sub match_internal ($$;@) {
-	return match_glob($_[0], $_[1], @_, internal => 1)
+	return match_glob(shift, shift, @_, internal => 1)
+}
+
+sub match_page ($$;@) {
+	my $page=shift;
+	my $match=match_glob($page, shift, @_);
+	if ($match && ! (exists $IkiWiki::pagesources{$page}
+	    && defined IkiWiki::pagetype($IkiWiki::pagesources{$page}))) {
+		return IkiWiki::FailReason->new("$page is not a page");
+	}
+	else {
+		return $match;
+	}
 }
 
 sub match_link ($$;@) {
@@ -2174,26 +2449,36 @@ sub match_link ($$;@) {
 
 	$link=derel($link, $params{location});
 	my $from=exists $params{location} ? $params{location} : '';
+	my $linktype=$params{linktype};
+	my $qualifier='';
+	if (defined $linktype) {
+		$qualifier=" with type $linktype";
+	}
 
 	my $links = $IkiWiki::links{$page};
-	return IkiWiki::FailReason->new("$page has no links", "" => 1)
+	return IkiWiki::FailReason->new("$page has no links", $page => $IkiWiki::DEPEND_LINKS, "" => 1)
 		unless $links && @{$links};
 	my $bestlink = IkiWiki::bestlink($from, $link);
 	foreach my $p (@{$links}) {
+		next unless (! defined $linktype || exists $IkiWiki::typedlinks{$page}{$linktype}{$p});
+
 		if (length $bestlink) {
-			return IkiWiki::SuccessReason->new("$page links to $link", $page => $IkiWiki::DEPEND_LINKS, "" => 1)
-				if $bestlink eq IkiWiki::bestlink($page, $p);
+			if ($bestlink eq IkiWiki::bestlink($page, $p)) {
+				return IkiWiki::SuccessReason->new("$page links to $link$qualifier", $page => $IkiWiki::DEPEND_LINKS, "" => 1)
+			}
 		}
 		else {
-			return IkiWiki::SuccessReason->new("$page links to page $p matching $link", $page => $IkiWiki::DEPEND_LINKS, "" => 1)
-				if match_glob($p, $link, %params);
+			if (match_glob($p, $link, %params)) {
+				return IkiWiki::SuccessReason->new("$page links to page $p$qualifier, matching $link", $page => $IkiWiki::DEPEND_LINKS, "" => 1)
+			}
 			my ($p_rel)=$p=~/^\/?(.*)/;
 			$link=~s/^\///;
-			return IkiWiki::SuccessReason->new("$page links to page $p_rel matching $link", $page => $IkiWiki::DEPEND_LINKS, "" => 1)
-				if match_glob($p_rel, $link, %params);
+			if (match_glob($p_rel, $link, %params)) {
+				return IkiWiki::SuccessReason->new("$page links to page $p_rel$qualifier, matching $link", $page => $IkiWiki::DEPEND_LINKS, "" => 1)
+			}
 		}
 	}
-	return IkiWiki::FailReason->new("$page does not link to $link", "" => 1);
+	return IkiWiki::FailReason->new("$page does not link to $link$qualifier", $page => $IkiWiki::DEPEND_LINKS, "" => 1);
 }
 
 sub match_backlink ($$;@) {
@@ -2243,7 +2528,7 @@ sub match_created_after ($$;@) {
 }
 
 sub match_creation_day ($$;@) {
-	if ((gmtime($IkiWiki::pagectime{shift()}))[3] == shift) {
+	if ((localtime($IkiWiki::pagectime{shift()}))[3] == shift) {
 		return IkiWiki::SuccessReason->new('creation_day matched');
 	}
 	else {
@@ -2252,7 +2537,7 @@ sub match_creation_day ($$;@) {
 }
 
 sub match_creation_month ($$;@) {
-	if ((gmtime($IkiWiki::pagectime{shift()}))[4] + 1 == shift) {
+	if ((localtime($IkiWiki::pagectime{shift()}))[4] + 1 == shift) {
 		return IkiWiki::SuccessReason->new('creation_month matched');
 	}
 	else {
@@ -2261,7 +2546,7 @@ sub match_creation_month ($$;@) {
 }
 
 sub match_creation_year ($$;@) {
-	if ((gmtime($IkiWiki::pagectime{shift()}))[5] + 1900 == shift) {
+	if ((localtime($IkiWiki::pagectime{shift()}))[5] + 1900 == shift) {
 		return IkiWiki::SuccessReason->new('creation_year matched');
 	}
 	else {
@@ -2274,11 +2559,13 @@ sub match_user ($$;@) {
 	my $user=shift;
 	my %params=@_;
 	
+	my $regexp=IkiWiki::glob2re($user);
+	
 	if (! exists $params{user}) {
 		return IkiWiki::ErrorReason->new("no user specified");
 	}
 
-	if (defined $params{user} && lc $params{user} eq lc $user) {
+	if (defined $params{user} && $params{user}=~/^$regexp$/i) {
 		return IkiWiki::SuccessReason->new("user is $user");
 	}
 	elsif (! defined $params{user}) {
@@ -2326,4 +2613,22 @@ sub match_ip ($$;@) {
 	}
 }
 
+package IkiWiki::SortSpec;
+
+# This is in the SortSpec namespace so that the $a and $b that sort() uses
+# are easily available in this namespace, for cmp functions to use them.
+sub sort_pages {
+	my $f=shift;
+	sort $f @_
+}
+
+sub cmp_title {
+	IkiWiki::pagetitle(IkiWiki::basename($a))
+	cmp
+	IkiWiki::pagetitle(IkiWiki::basename($b))
+}
+
+sub cmp_mtime { $IkiWiki::pagemtime{$b} <=> $IkiWiki::pagemtime{$a} }
+sub cmp_age { $IkiWiki::pagectime{$b} <=> $IkiWiki::pagectime{$a} }
+
 1
